"""
LeMMA - the Lazy man's Editor for MMA
Note that this is NOT really an MMA editor, but rather a simple front-end to MMA

This is the main application module

Software is provided AS IS. Feel free to use AT YOUR OWN RISK!

Version 0.6. Copyright (c) Gek S. Low. 9 Feb 2008
"""


# app.py

from Tkinter import *
import tkMessageBox
import tkFileDialog
import re
import cPickle

import settings
from settings import AutoScrollbar
from constants import *

version = "0.6a (29 Feb 2008)"
MMA_filetypes = (("MMA files", "*.mma"), ("All files","*.*"))

def blah():
	tkMessageBox.showinfo("Under construction", "This feature is not yet implemented")

class Application(Frame):
	def __init__(self, master=None):
		Frame.__init__(self, master)
		self.grid(sticky=N+S+E+W)
		self.option_add("*font", DEFAULT_FONT)
		self.createWidgets()
		settings.readGrooves()
		settings.readSettings()
		settings.determineLibDir()

	def createWidgets(self):
		# Make top-level window resizable
		top=self.winfo_toplevel()
		top.rowconfigure(0, weight=1)
		top.columnconfigure(0, weight=1)
		self.rowconfigure(2, weight=1)
		self.columnconfigure(0, weight=1)

		# Main toolbar
		self.BtnFrame = Frame(self)
		self.BtnFrame.grid(row=0, column=0, sticky=NW)
		
		self.newBtn = Button(self.BtnFrame, text="New", command=self.newFile)
		self.newBtn.grid(row=0, column=0, sticky=NW)
		self.loadMMABtn = Button(self.BtnFrame, text="Load MMA", command=self.loadMMA)
		self.loadMMABtn.grid(row=0, column=1, sticky=NW)
		self.saveMMABtn = Button(self.BtnFrame, text="Save MMA", command=self.saveAsMMA)
		self.saveMMABtn.grid(row=0, column=3, sticky=NW)
		self.viewMMABtn = Button(self.BtnFrame, text="View MMA", command=self.viewMMA)
		self.viewMMABtn.grid(row=0, column=4, sticky=NW)
		self.playBtn = Button(self.BtnFrame, text="Play", command=self.playFile)
		self.playBtn.grid(row=0, column=5, sticky=NW)
		self.settingsBtn = Button(self.BtnFrame, text="Settings", command=self.settings)
		self.settingsBtn.grid(row=0, column=6, sticky=NW)
		self.aboutBtn = Button(self.BtnFrame, text="About", command=self.about)
		self.aboutBtn.grid(row=0, column=7, sticky=NW)
		self.helpBtn = Button(self.BtnFrame, text="Help", command=self.help)
		self.helpBtn.grid(row=0, column=8, sticky=NW)
		self.quitBtn = Button(self.BtnFrame, text="Quit", command=self.quit)
		self.quitBtn.grid(row=0, column=9, sticky=NW)
				
		# Time Signature, Key Signature, Tempo
		self.infoFrame = Frame(self)
		self.infoFrame.grid(row=1, column=0, sticky=NW, padx=5, pady=5)
		
		Label(self.infoFrame, text="Time Sig ").grid(row=0, column=0, sticky=NW)
		self.timeSigEntry = Entry(self.infoFrame, width=4)
		self.timeSigEntry.insert(1,"4/4")
		self.timeSigEntry.grid(row=0, column=1, sticky=NW)
		
		Label(self.infoFrame, text="   Key ").grid(row=0, column=2, sticky=NW)
		self.keyEntry = Entry(self.infoFrame, width=6)
		self.keyEntry.insert(1,"C")
		self.keyEntry.grid(row=0, column=3, sticky=NW)
		
		Label(self.infoFrame, text="   Tempo ").grid(row=0, column=4, sticky=NW)
		self.tempoEntry = Entry(self.infoFrame, width=4)
		self.tempoEntry.insert(1,"60")
		self.tempoEntry.grid(row=0, column=5, sticky=NW)
		
		self.SwingMode = 1
		self.swingCheckBtn = Checkbutton(self.infoFrame, text="Swing Mode", command=self.toggleSwingMode)
		self.swingCheckBtn.grid(row=0, column=6, sticky=NW)
		self.swingCheckBtn.select()
		
		# This is where the chords are entered
		self.mainCanvas = Canvas(self, background=MAINFRAME_COLOR_BG, highlightthickness=0)
		self.mainCanvas.grid(row=2, column=0, sticky=N+S+E+W, padx=5, pady=5)
		self.mainFrame = Frame(self.mainCanvas, background=MAINFRAME_COLOR_BG, borderwidth=0)
		#self.mainFrame.grid(row=2, column=0, sticky=NW, padx=5, pady=5)
		
		self.scrollY = AutoScrollbar(self, orient=VERTICAL, command=self.mainCanvas.yview, takefocus=0)
		self.scrollY.grid(row=2, column=1, stick=N+S)
		self.scrollX = AutoScrollbar(self, orient=HORIZONTAL, command=self.mainCanvas.xview, takefocus=0)
		self.scrollX.grid(row=3, column=0, sticky=E+W)

		self.mainCanvas["xscrollcommand"] = self.scrollX.set
		self.mainCanvas["yscrollcommand"] = self.scrollY.set
		self.scrollX["command"] = self.mainCanvas.xview
		self.scrollY["command"] = self.mainCanvas.yview

		self.measures = []
		self.grooves = []
		self.grooves_lib = []
		# this is used for SetLibPath directive
		self.grooves_libpath = []
		# this is used for SetAutoLibPath directive
		self.grooves_autolibpath = []
		self.barlines = []
		self.codebtns = []
		self.codes = []
		# See constants.py for default settings. Adjust to taste.
		for y in range(NUM_MEASURES_Y):
			for x in range(NUM_MEASURES_X):
				c = x + y * NUM_MEASURES_X
				self.measures += [Measure(self.mainFrame, adjustCanvas=self.adjustCanvasScroll, id=c)]
				Label(self.mainFrame, text=str(c+1), justify=RIGHT, background=BARNUM_COLOR_BG, foreground=BARNUM_COLOR_FG, takefocus=0, font=BARNUM_FONT).grid(row=y*2, column=x*3)
				def setGroove(index=c):
					if self.grooves[index]["text"] != GROOVE_TEXT_NONE:
						if not self.grooves[index]["text"]+"|"+self.grooves_libpath[index] in settings.groovelib_lookup.keys():
							tkMessageBox.showwarning("Warning", "Groove "+self.grooves[index]["text"]+" is not found in LeMMA groove database")
							settings.currentgroovelib = ""
						else:
							settings.currentgroovelib = self.grooves_lib[index] + "|" + self.grooves_libpath[index]
						d = settings.GroovesDialog(self, self.grooves[index]["text"])
					else:
						d = settings.GroovesDialog(self, "")
					if d.selectedGroove != "":
						self.grooves[index].configure(text=d.selectedGroove, foreground=GROOVE_COLOR_FG, font=GROOVE_FONT)
						self.grooves_lib[index] = d.selectedGroove_lib
						self.grooves_libpath[index] = d.selectedGroove_libpath
						self.grooves_autolibpath[index] = d.selectedGroove_autolibpath
						#print d.selectedGroove, d.selectedGroove_lib, d.selectedGroove_libpath, d.selectedGroove_autolibpath
					else:
						self.grooves[index].configure(text=GROOVE_TEXT_NONE, foreground=GROOVE_COLOR_FG_NONE, font=GROOVE_FONT_NONE)
						self.grooves_lib[index] = ""
						self.grooves_libpath[index] = ""
						self.grooves_autolibpath[index] = ""

				self.grooves += [Button(self.mainFrame, text=GROOVE_TEXT_NONE, foreground=GROOVE_COLOR_FG_NONE, background=GROOVE_COLOR_BG, activebackground=GROOVE_COLOR_BG_MOUSEENTER, anchor=W, justify=LEFT, relief=FLAT, takefocus=0, command=setGroove, font=GROOVE_FONT_NONE, padx=0, pady=0, borderwidth=0)]
				self.grooves_lib += [""]
				self.grooves_libpath += [""]
				self.grooves_autolibpath += [""]

				def setCode(index=c):
					d = settings.CodeDialog(self, code=self.codes[index])
					self.codes[index] = d.code.strip()
					if self.codes[index] != "":
						self.codebtns[index].configure(foreground=CODE_COLOR_BG, background=CODE_COLOR_FG, font=CODE_FONT)
					else:
						self.codebtns[index].configure(foreground=CODE_COLOR_FG_NONE, background=CODE_COLOR_BG, font=CODE_FONT_NONE)

				self.codes += [""]
				self.codebtns += [Button(self.mainFrame, text=CODE_TEXT, foreground=CODE_COLOR_FG_NONE, background=CODE_COLOR_BG, activebackground=CODE_COLOR_BG_MOUSEENTER, anchor=W, justify=LEFT, relief=FLAT, takefocus=0, command=setCode, font=CODE_FONT_NONE, padx=0, pady=0, borderwidth=0)]
					
				def toggleBarlines(index=c):
					barline = self.barlines[index]["text"]
					if barline == " | ":
						self.barlines[index].configure(text=" |:")
					elif barline == " |:":
						self.barlines[index].configure(text=":| ")
					elif barline == ":| ":
						self.barlines[index].configure(text=":|:")
					else:
						self.barlines[index].configure(text=" | ")
				self.barlines += [Button(self.mainFrame, text=" | ", justify=CENTER, takefocus=0, background=BARLINE_COLOR_BG, foreground=BARLINE_COLOR_FG, activebackground=BARLINE_COLOR_BG_MOUSEENTER, anchor=CENTER, relief=FLAT, font=BARLINE_FONT, padx=0, pady=0, borderwidth=0, command=toggleBarlines)]

				self.barlines[c].grid(row=y*2+1, column=x*3, sticky=W)
				self.measures[c].grid(row=y*2+1, column=x*3+1, sticky=W, columnspan=2)
				self.grooves[c].grid(row=y*2, column=x*3+1, sticky=W)
				self.codebtns[c].grid(row=y*2, column=x*3+2, sticky=E)

		# Focus on the first measure and set up the main canvas
		self.measures[0].contents.focus()
		self.mainWindow = self.mainCanvas.create_window(0, 0, window=self.mainFrame, anchor=NW)
		self.adjustCanvasScroll(self.measures[0])

		self.outputText = Text(self, width=70, height=5, bg="white", wrap=WORD, relief=GROOVE, takefocus=0, font=TEXT_FONT)
		self.outputText.grid(row=4, column=0, sticky=S+W+E, columnspan=2)
		
		settings.outputWindow = self.outputText

	def adjustCanvasScroll(self, measure):
		# Adjusts the size of scrolling area
		self.mainFrame.update_idletasks()
		self.mainCanvas.configure(scrollregion=self.mainCanvas.bbox(ALL))
		# need to scroll to the current measure?
		# Y scroll
		y1 = measure.winfo_y() - self.grooves[0].winfo_height()
		y2 = measure.winfo_y() + measure.winfo_height()
		h = self.mainFrame.winfo_height()
		f1 = float(y1) / h
		f2 = float(y2) / h
		(s1, s2) = self.scrollY.get()
		#print f1, f2, self.scrollY.get()
		if f1 < s1:
			self.mainCanvas.yview_moveto(f1)
		if f2 > s2:
			self.mainCanvas.yview_moveto(f2)
		
		# X scroll
		x1 = measure.winfo_x() - self.barlines[0].winfo_width()
		x2 = measure.winfo_x() + max(measure.winfo_width(), self.grooves[measure.id].winfo_width() + self.codebtns[measure.id].winfo_width())
		w = self.mainFrame.winfo_width()
		f1 = float(x1) / w
		f2 = float(x2) / w
		(s1, s2) = self.scrollX.get()
		#print f1, f2, self.scrollY.get()
		if f1 < s1:
			self.mainCanvas.xview_moveto(f1)
		if f2 > s2:
			self.mainCanvas.xview_moveto(f2)


	def toggleSwingMode(self):
		self.SwingMode = not self.SwingMode
	
	def newFile(self):
		if tkMessageBox.askyesno("Warning", "All contents will be lost! Continue?"):
			self.clearCanvas()

	def clearCanvas(self):
		for y in range(NUM_MEASURES_Y):
			for x in range(NUM_MEASURES_X):
				c = x + y * NUM_MEASURES_X
				self.measures[c].contents.delete(0, END)
				self.measures[c].contents.configure(width=DEFAULT_MEASURE_WIDTH)
				self.grooves[c].configure(text=GROOVE_TEXT_NONE, foreground=GROOVE_COLOR_FG_NONE, font=GROOVE_FONT_NONE)
				self.grooves_lib[c] = ""
				self.grooves_libpath[c] = ""
				self.grooves_autolibpath[c] = ""
				self.barlines[c].configure(text=" | ")
				self.codes[c] = ""
				self.codebtns[c].configure(foreground=CODE_COLOR_FG_NONE, background=CODE_COLOR_BG, font=CODE_FONT_NONE)

		# Update the canvas scroll area
		self.measures[0].adjustCanvas(self.measures[0])


	def loadMMA(self):
		currentLibPath = settings.libDir
		currentAutoLibPath = "stdlib"

		global MMA_filetypes
		self.currentFile = tkFileDialog.askopenfilename(filetypes = MMA_filetypes)
		
		if self.currentFile == "" or self.currentFile == ():
			return
		
		self.clearCanvas()
		f = open(self.currentFile, "r")
		i = 0 # measure number
		
		measure = 0
		repeatending_count = 1
		clearcontents = 1
		reached_first_measure = False
		
		# read each line
		for line in f:
			line = line.strip()
			matched = False
			
			if reached_first_measure == False and re.compile(r'KeySig\s*').match(line):
				line = re.compile(r'KeySig\s*').sub("", line)
				self.keyEntry.delete(0, END)
				self.keyEntry.insert(0, line)
				continue
			
			if reached_first_measure == False and re.compile(r'TimeSig\s*').match(line):
				line = re.compile(r'TimeSig\s*').sub("", line)
				line = line.replace(" " , "/")
				self.timeSigEntry.delete(0, END)
				self.timeSigEntry.insert(0, line)
				continue
			
			if reached_first_measure == False and re.compile(r'Tempo\s*').match(line):
				line = re.compile(r'Tempo\s*').sub("", line)
				self.tempoEntry.delete(0, END)
				self.tempoEntry.insert(0, line)
				continue
			
			if reached_first_measure == False and re.compile(r'SwingMode\s*').match(line):
				line = re.compile(r'SwingMode\s*').sub("", line)
				if line == "On":
					self.swingCheckBtn.select()
				else:
					self.swingCheckBtn.deselect()
				continue

			if re.compile(r'SetLibPath\s*').match(line):
				if not reached_first_measure:
					reached_first_measure = True
				line = re.compile(r'SetLibPath\s*').sub("", line)
				self.grooves_libpath[measure] = line
				currentLibPath = line
				continue
	
			if re.compile(r'SetAutoLibPath\s*').match(line):
				if not reached_first_measure:
					reached_first_measure = True
				line = re.compile(r'SetAutoLibPath\s*').sub("", line)
				self.grooves_autolibpath[measure] = line
				currentAutoLibPath = line
				continue
			
			if re.compile(r'Groove\s*').match(line):
				if not reached_first_measure:
					reached_first_measure = True
				line = re.compile(r'Groove\s*').sub("", line)
				self.grooves[measure].configure(text=line, foreground=GROOVE_COLOR_FG, font=GROOVE_FONT)
				if self.grooves_libpath[measure] == "":
					self.grooves_libpath[measure] = currentLibPath
				if self.grooves_autolibpath[measure] == "":
					self.grooves_autolibpath[measure] = currentAutoLibPath
				if line+"|"+self.grooves_libpath[measure] in settings.groovelib_lookup.keys():
					self.grooves_lib[measure] = settings.groovelib_lookup[line+"|"+self.grooves_libpath[measure]]
				else:
					tkMessageBox.showwarning("Warning", "Groove "+line+" is not found in LeMMA groove database")
				continue

			if clearcontents == 1:
				contents = ""
			
			# is this a repeat, repeatending?
			if line == "Repeat":
				#contents = "|: "
				contents = ""
				matched = True
				# Is this a repeatend + repeat?
				if self.barlines[measure].configure(text=":| "):
					self.barlines[measure].configure(text=":|:")
				else:
					self.barlines[measure].configure(text=" |:")
				repeatending_count = 1
				clearcontents = 0

			if line == "RepeatEnding":
				contents = "[" + str(repeatending_count) + " "
				matched = True
				repeatending_count += 1
				clearcontents = 0
				
			# form the measure contents field
			if re.compile(r'\d+').match(line):
				line = re.compile(r'^\d+\s*').sub("", line)
				contents = contents + line
				matched = True
				clearcontents = 1
			
			# is this a repeat end? then need to retrieve and add to previous bar
			if line == "RepeatEnd":
				self.barlines[measure].configure(text=":| ")
				#measure = measure - 1
				#contents = self.measures[measure].contents.get()
				#self.measures[measure].contents.delete(0, END)
				#contents += " :|"
				clearcontents = 1
				matched = True
			
			# insert into measure
			if clearcontents == 1 and contents != "":
				self.measures[measure].contents.insert(0, contents)
				measure = measure + 1
				if measure > len(self.measures):
					break
				continue
			
			if line == "cut -1":
				break

			# Must be some MMA code
			if reached_first_measure and not matched:
				#print measure, line
				self.codes[measure] = self.codes[measure] + line
				self.codebtns[measure].configure(foreground=CODE_COLOR_BG, background=CODE_COLOR_FG, font=CODE_FONT)

		# Update the canvas scroll area
		self.measures[0].adjustCanvas(self.measures[0])		
		f.close()
		settings.printOutput("Loaded " + self.currentFile)

	def viewMMA(self):
		self.saveMMA("_temp_.mma")
		settings.ViewFileDialog(self, title="View MMA", filename="_temp_.mma")
	
	def playFile(self):
		self.saveMMA("_temp_.mma")
		settings.playMMA()
		
	def saveAsMMA(self):
		self.currentFile = tkFileDialog.asksaveasfilename(filetypes = MMA_filetypes)
		if self.currentFile != "" and self.currentFile != ():
			self.saveMMA(self.currentFile)
			settings.printOutput("Saved " + self.currentFile)
			return
		settings.printOutput("Save cancelled")
		
	def saveMMA(self, filename):
		currentLibPath = settings.libDir
		currentAutoLibPath = "stdlib"

		f = open(filename, "w")
		i = 0	# measure number
		print >>f, "// MMA file generated by LeMMA version " + version + "\n"
		print >>f, "KeySig " + self.keyEntry.get()
		print >>f, "TimeSig " + self.timeSigEntry.get().replace("/", " ")
		print >>f, "Tempo " + self.tempoEntry.get() + "\n"
		if self.SwingMode == 1:
			print >>f, "SwingMode On\n"
		else:
			print >>f, "SwingMode Off\n"
			
		#swingmode?
		reRepeatEnding = re.compile(r'\[\d+')

		for measure in self.measures:
			groove = self.grooves[i]["text"]
			groove_libpath = self.grooves_libpath[i]
			groove_autolibpath = self.grooves_autolibpath[i]
			
			# set default groove (Folk) if first measure is empty
			if i == 0 and groove == GROOVE_TEXT_NONE:
				groove = "Folk"
				groove_libpath = settings.libDir
				groove_autolibpath = "stdlib"
		
			contents = measure.contents.get()
			
			isRepeat = 0
			isRepeatEnd = 0
			isRepeatEnding = 0
			
			# Legacy stuff, if user types the repeat signs directly into measure text
			if contents.find("|:") != -1:
				isRepeat = 1
				contents = contents.replace("|:","")
				
			if contents.find(":|") != -1:
				isRepeatEnd = 1
				contents = contents.replace(":|","")
				
			if reRepeatEnding.match(contents):
				isRepeatEnding = 1
				contents = reRepeatEnding.sub("", contents) # first repeat
			
			contents = contents.strip(' ')
			if contents != "":
				printRepeat = False
				if self.barlines[i]["text"] == ":| ":
					print >>f, "RepeatEnd"
				if self.barlines[i]["text"] == ":|:":
					print >>f, "RepeatEnd"
					printRepeat = True

				# print the groove first, if any
				if groove != GROOVE_TEXT_NONE:
					if groove_libpath != currentLibPath:
						print >>f, "SetLibPath " + groove_libpath
						currentLibPath = groove_libpath
					if groove_autolibpath != currentAutoLibPath:
						print >>f, "SetAutoLibPath " + groove_autolibpath
						currentAutoLibPath = groove_autolibpath
					print >>f, "Groove " + groove

				# insert MMA codes
				if self.codes[i] != "":
					print >>f, self.codes[i]

				# now for the measure
				# new barline approach to repeats
				if printRepeat or self.barlines[i]["text"] == " |:":
					print >>f, "Repeat"

				if isRepeat == 1:
					print >>f, "Repeat"
				if isRepeatEnding == 1:
					print >>f, "RepeatEnding"
				
				print >>f, str(i+1) + " " + contents
				
				if isRepeatEnd == 1:
					print >>f, "RepeatEnd"
			else:
				break	# stops at first empty measure, ignoring any other non-empty measures after it
			i = i+1
			
		print >>f, "cut -1"
		f.close()
		
	def about(self):
		tkMessageBox.showinfo("About LeMMA",
			"LeMMA - the Lazyman's editor for MMA\n\n"
			"Version " + version + "\n\n"
			"Copyright (c) Gek S. Low")
	
	def help(self):
		d = settings.ViewFileDialog(self, title="Help", filename="help.txt")

	def settings(self):
		d = settings.SettingsDialog(self)

class Measure(Frame):
	def __init__(self, master=None, adjustCanvas=None, id=0):
		self.adjustCanvas = adjustCanvas
		self.id = id
		Frame.__init__(self, master, borderwidth=0, bg=MAINFRAME_COLOR_BG)
		self.grid()
		self.createWidgets()
	
	def createWidgets(self):
		self.measurewidth=DEFAULT_MEASURE_WIDTH
		self.contents = Entry(self, width=DEFAULT_MEASURE_WIDTH, background=MEASURE_COLOR_BG, foreground=MEASURE_COLOR_FG, borderwidth=0, highlightcolor=MEASURE_COLOR_BG_FOCUS, highlightthickness=1, relief=FLAT, font=MEASURE_FONT, justify=LEFT)
		self.contents.grid(row=0,column=0)
			
		self.contents.bind("<FocusOut>", self.doFocusOut)
		self.contents.bind("<FocusIn>", self.doFocusIn)
		self.contents.bind("<Enter>", self.doEnter)
		self.contents.bind("<Leave>", self.doLeave)

	def doEnter(self, event):
		self.contents.configure(highlightbackground=MEASURE_COLOR_HIGHLIGHT_MOUSEENTER)

	def doLeave(self, event):
		self.contents.configure(highlightbackground=MEASURE_COLOR_HIGHLIGHT)

	def doFocusIn(self, event):
		self.contents.configure(background=MEASURE_COLOR_BG_FOCUS)
		# need to scroll the canvas?
		self.adjustCanvas(self)

	def doFocusOut(self, event):
		# set background back to white
		self.contents.configure(background=MEASURE_COLOR_BG)

		# Adjust the width
		w = len(str(self.contents.get()))
		if w >= DEFAULT_MEASURE_WIDTH:
			self.contents.configure(width=w+2)
		else:
			self.contents.configure(width=DEFAULT_MEASURE_WIDTH)
		self.contents.index(1)
		if self.adjustCanvas != None and self.contents["width"] != self.measurewidth:
			self.measurewidth = w
			self.adjustCanvas(self)
